from enum import Enum


SIDE_BUY = 'BUY'
SIDE_SELL = 'SELL'

ORDER_TYPE_LIMIT = 'LIMIT'
ORDER_TYPE_MARKET = 'MARKET'
ORDER_TYPE_STOP_LOSS = 'STOP_LOSS'
ORDER_TYPE_STOP_LOSS_LIMIT = 'STOP_LOSS_LIMIT'
ORDER_TYPE_TAKE_PROFIT = 'TAKE_PROFIT'
ORDER_TYPE_TAKE_PROFIT_LIMIT = 'TAKE_PROFIT_LIMIT'
ORDER_TYPE_LIMIT_MAKER = 'LIMIT_MAKER'


class Resolution(Enum):
    ONE_HOUR = '60'
    THREE_HOURS = '180'
    SIX_HOURS = '360'
    TWELVE_HOURS = '720'
    ONE_DAY = 'D'
    TWO_DAYS = '2D'
    THREE_DAYS = '3D'



from typing import Union
import inspect

import requests
from simplejson import JSONDecodeError
import json

from warnings import warn

from deprecation import deprecated

from .exceptions import *
from ._package_data import __version__

DEPRECATED_IN = __version__
CURRENT_VERSION = __version__
REMOVE_IN = '2.0.0'
DEPRECATION_MESSAGE = 'This method is deprecated. Please start using latest version Wrapper. ' \
                      'And do it faster, Im not even sure if I deprecated it correctly. '


def validate_response(response: dict) -> bool:
    return response.get('success') or str(response.get('success')).lower() == 'true'


@deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
            current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
def get_token(email: str, password: str, remember: bool = False) -> str:
    _ = locals()
    warn(DEPRECATION_MESSAGE, DeprecationWarning)
    func_name = inspect.currentframe().f_code.co_name

    payload = json.dumps({
        "email": email,
        "password": password,
        'remember': remember
    })
    headers = {
        'Content-Type': 'application/json'
    }

    try:
        r = requests.post('https://api.wallex.ir/auth/login/email', headers=headers, data=payload, timeout=5)
    except Exception as e:
        raise RequestsExceptions(func_name, e, _)

    status_code = r.status_code

    if 200 <= status_code < 300:
        try:
            resp = r.json()
        except JSONDecodeError as e:
            raise WallexExceptions(func_name, r.text, _)

        if validate_response(resp):
            return resp.get('result').get('token')
        else:
            raise WallexExceptions(func_name, r.text, _)
    else:
        raise StatusCodeError(func_name, str(status_code), r.text, _)


class Wallex:
    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def __init__(self, token: str, timeout: int = 5, verify: Union[bool, None] = None):
        self.base_url = "https://api.wallex.ir/v1/"
        self.verify = verify
        self.timeout = timeout
        self.session = requests.Session()
        self.session.headers = {
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {token}',
            'Accept': 'application/json',
        }

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def all_market_stats(self):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name
        try:
            r = self.session.get(self.base_url + 'markets',
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result')
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def symbol_market_stats(self, symbol: str):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name
        try:
            r = self.session.get(self.base_url + 'markets',
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result').get('symbols').get(symbol.upper())
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def order_book(self, symbol: str):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name
        try:
            r = self.session.get(self.base_url + f'depth?symbol={symbol.upper()}',
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result')
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def all_balances(self):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name
        try:
            r = self.session.get(self.base_url + 'account/balances',
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result').get('balances')
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def coin_balance(self, coin: str):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        coin = coin.upper()
        func_name = inspect.currentframe().f_code.co_name
        try:
            r = self.session.get(self.base_url + 'account/balances',
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result').get('balances').get(coin)
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def coin_available_balance(self, coin: str):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        coin = coin.upper()
        func_name = inspect.currentframe().f_code.co_name
        try:
            r = self.session.get(self.base_url + 'account/balances',
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                _ = resp.get('result').get('balances').get(coin)
                return float(_.get('value')) - float(_.get('locked'))
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def create_order(self, price: str, quantity: str, side: str, symbol: str, type_: str, client_id: str = None):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name

        try:
            payload = {
                "price": price,
                "quantity": quantity,
                "side": side.lower(),
                "symbol": symbol.upper(),
                "type": type_.lower(),
            }
            if client_id:
                payload.update({'client_id': client_id})
            payload = json.dumps(payload)

            r = self.session.post(self.base_url + 'account/orders', data=payload,
                                  timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result')
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def cancel_order(self, client_id: str):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name

        try:
            payload = json.dumps({
                "clientOrderId": client_id
            })
            r = self.session.delete(self.base_url + 'account/orders', data=payload,
                                    timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result')
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def open_orders(self, symbol: str):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name

        try:
            r = self.session.get(self.base_url + f'account/openOrders?symbol={symbol.upper()}',
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result').get('orders')
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def open_orders_by_side(self, symbol: str, side: str):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name

        try:
            r = self.session.get(self.base_url + f'account/openOrders?symbol={symbol.upper()}',
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                all_orders = resp.get('result').get('orders')
                __ = []
                for order in all_orders:
                    order: dict
                    if order.get('side').lower() == side.lower():
                        __.append(order)
                return __

            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def user_recent_trades(self, symbol: str = None, side: str = None, active: bool = None):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name

        params = {}

        if symbol:
            params.update({'symbol': symbol.upper()})
        if side:
            params.update({'side': side.lower()})
        if active:
            params.update({'active': active})

        try:
            r = self.session.get(self.base_url + f'account/trades', params=params,
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result').get('AccountLatestTrades')
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def get_order_by_id(self, order_id: str):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name

        try:
            r = self.session.get(self.base_url + f'account/orders/{order_id}',
                                 timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result')
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)

    @deprecated(deprecated_in=DEPRECATED_IN, removed_in=REMOVE_IN,
                current_version=CURRENT_VERSION, details=DEPRECATION_MESSAGE)
    def withdraw(self, coin: str, client_unique_id: str, network: str, amount: str, address: str, memo: str = None):
        _ = locals()
        warn(DEPRECATION_MESSAGE, DeprecationWarning)
        func_name = inspect.currentframe().f_code.co_name

        try:
            payload = {
                "coin": coin.upper(),
                "client_unique_id": str(client_unique_id),
                "network": network,
                "value": str(amount),
                "wallet_address": address,
            }
            if memo:
                payload.update({'memo': memo})
            payload = json.dumps(payload)
            r = self.session.post(self.base_url + f'account/crypto-withdrawal', data=payload,
                                  timeout=self.timeout, verify=self.verify)
        except Exception as e:
            raise RequestsExceptions(func_name, e, _)

        status_code = r.status_code

        if 200 <= status_code < 300:
            try:
                resp = r.json()
            except JSONDecodeError as e:
                raise WallexExceptions(func_name, r.text, _)

            if validate_response(resp):
                return resp.get('result')
            else:
                raise WallexExceptions(func_name, r.text, _)

        else:
            raise StatusCodeError(func_name, str(status_code), r.text, _)



import typing as t
from requests import Response


__all__ = [
    'WallexExceptions',
    'RequestsExceptions',
    'RequestTimeout',
    'TokenException',
    'InvalidResponse',
    'StatusCodeError',
    'JSONDecodingError',
    'InvalidInputs',
    'WebsocketExceptions',
    'WebsocketNotConnected'
]


class WallexExceptions(Exception):
    def __init__(
            self, func_name: str, message: t.Union[str, t.Type[Exception], Exception], *args, **kwargs
    ):
        """
        Base exception class for Bitpin.

        :param func_name: function name that raise exception
        :type func_name: str

        :param message: message of exception
        :type message: str | Exception

        :param args: args of exception
        :type args: t.Any

        :param kwargs: kwargs of exception
        :type kwargs: t.Any

        :return: None
        :rtype: None
        """

        self.func_name = func_name
        self.message = str(message)
        self.f_args = args
        self.f_kwargs = kwargs
        super().__init__(self.message)

    def __str__(self):
        __str = f'[{self.func_name}] -> {self.message}'

        if self.f_args and len(self.f_args) > 0:
            __str += f' | Args: {self.f_args}'
        if self.f_kwargs:
            __str += f' | Kwargs: {self.f_kwargs}'

        return __str


class RequestsExceptions(WallexExceptions):
    """ Exception class for requests error. """


class RequestTimeout(RequestsExceptions):
    """ Exception class for requests timeout error. """


class TokenException(RequestsExceptions):
    """ Exception class for invalid token error. """


class ProcessExceptions(WallexExceptions):
    """ Exception class for process error. """

    def __init__(self, func_name: str, message: str, response: Response, *args, **kwargs):
        self.response = response
        msg = f'{message} || {response.text}'
        super().__init__(func_name, msg, *args, **kwargs)


class StatusCodeError(ProcessExceptions):
    """ Exception class for status code error. """


class JSONDecodingError(ProcessExceptions):
    """ Exception class for json decode error. """


class InvalidResponse(ProcessExceptions):
    """ Exception class for invalid response error. """


class InvalidInputs(WallexExceptions):
    """ Exception class for invalid requested data error. """


class WebsocketExceptions(WallexExceptions):
    """ Exception class for websocket error. """


class WebsocketNotConnected(WebsocketExceptions):
    """ Exception class for websocket not connected error. """



import typing as t
from abc import ABC, abstractmethod

import requests


from ..enums import Resolution


__all__ = [
    'BaseClient',
    # 'AbstractClient'
]


class BaseClient(ABC):
    API_URL = 'https://api.wallex.ir'
    PUBLIC_API_VERSION = 'v1'

    REQUEST_TIMEOUT: float = 10

    SIDE_BUY = 'BUY'
    SIDE_SELL = 'SELL'

    ORDER_TYPE_LIMIT = 'LIMIT'
    ORDER_TYPE_MARKET = 'MARKET'
    ORDER_TYPE_STOP_LOSS = 'STOP_LOSS'
    ORDER_TYPE_STOP_LOSS_LIMIT = 'STOP_LOSS_LIMIT'
    ORDER_TYPE_TAKE_PROFIT = 'TAKE_PROFIT'
    ORDER_TYPE_TAKE_PROFIT_LIMIT = 'TAKE_PROFIT_LIMIT'
    ORDER_TYPE_LIMIT_MAKER = 'LIMIT_MAKER'

    def __init__(
            self, api_key: t.Optional[str] = None, requests_params: t.Optional[t.Dict[str, str]] = None,
    ):
        self.API_KEY = api_key

        self._requests_params = requests_params
        self.session = self._init_session()

    @staticmethod
    def _get_kwargs(locals_: t.Dict, del_keys: t.List[str] = None, del_nones: bool = False) -> t.Dict:
        _del_keys = ['self', 'cls']
        if del_keys is not None:
            _del_keys.extend(del_keys)

        if del_nones is True:
            return {key: value for key, value in locals_.items() if (key not in _del_keys) and (value is not None)}

        return {key: value for key, value in locals_.items() if key not in _del_keys}

    @staticmethod
    def _get_headers() -> t.Dict:
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
        }
        return headers

    def _create_api_uri(self, path: str, version: str = PUBLIC_API_VERSION) -> str:
        return self.API_URL + '/' + version + '/' + path

    def _get_request_kwargs(self, method, signed: bool, **kwargs) -> t.Dict:
        # set default requests timeout
        kwargs['timeout'] = self.REQUEST_TIMEOUT

        # add our global requests params
        if self._requests_params:
            kwargs.update(self._requests_params)

        data = kwargs.get('data', None)
        if data and isinstance(data, dict):
            kwargs['data'] = data

            # find any requests params passed and apply them
            if 'requests_params' in kwargs['data']:
                # merge requests params into kwargs
                kwargs.update(kwargs['data']['requests_params'])
                del (kwargs['data']['requests_params'])

        if signed is True:
            headers = kwargs.get('headers', {})
            headers.update({'x-api-key': self.API_KEY})
            kwargs['headers'] = headers

        # if get request assign data array to params value for requests lib
        if data and method == 'get':
            kwargs['params'] = '&'.join('%s=%s' % (data[0], data[1]) for data in kwargs['data'])
            del (kwargs['data'])

        return kwargs

    @staticmethod
    def _pick(result: t.Union[t.Dict, t.List], key: str, value: str = None) -> t.Union[t.Dict, t.List]:
        result = result.copy()
        if isinstance(result, dict):
            symbol_data = result.get(key)
            result.clear()
            result[key] = symbol_data

            return result
        if isinstance(result, list):
            assert value is not None, "value is required for list"
            # print(f"Looking for {key=} with {value=} in {result}")
            result_ = [item for item in result if item[key] == value]

            return result_

    @abstractmethod
    def _init_session(self) -> requests.Session:
        raise NotImplementedError('_init_session not implemented')

    @abstractmethod
    def _request(self, method, uri: str, signed: bool, **kwargs):
        raise NotImplementedError('_request not implemented')

    @staticmethod
    @abstractmethod
    def _handle_response(response: requests.Response):
        raise NotImplementedError('_handle_response not implemented')

    @abstractmethod
    def _request_api(
            self, method, path: str, signed: bool = False, version=PUBLIC_API_VERSION, **kwargs
    ):
        raise NotImplementedError('_request_api not implemented')

    @abstractmethod
    def _get(self, path, signed=False, version=PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        raise NotImplementedError('_get not implemented')

    @abstractmethod
    def _post(self, path, signed=False, version=PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        raise NotImplementedError('_post not implemented')

    @abstractmethod
    def _put(self, path, signed=False, version=PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        raise NotImplementedError('_put not implemented')

    @abstractmethod
    def _delete(self, path, signed=False, version=PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        raise NotImplementedError('_delete not implemented')

    @abstractmethod
    def get_market_stats(self, symbol: str = None) -> t.Dict:
        raise NotImplementedError('get_market_stats not implemented')

    @abstractmethod
    def get_currencies(self, currency: str = None) -> t.Dict:
        raise NotImplementedError('get_currencies not implemented')

    @abstractmethod
    def get_currencies_stats(self, currency=None) -> t.Dict:
        raise NotImplementedError('get_currencies_stats not implemented')

    @abstractmethod
    def get_orderbook(self, symbol: str) -> t.Dict:
        raise NotImplementedError('get_orderbook not implemented')

    @abstractmethod
    def get_recent_trades(self, symbol: str = 'None', **kwargs) -> t.Dict:
        raise NotImplementedError('get_recent_trades not implemented')

    @abstractmethod
    def get_ohlc_data(
            self, symbol: str = None, resolution: Resolution = None, from_date: int = None, to_date: int = None
    ) -> t.Dict:
        raise NotImplementedError('get_ohlc_data not implemented')

    @abstractmethod
    def get_profile(self) -> t.Dict:
        raise NotImplementedError('get_profile not implemented')

    @abstractmethod
    def get_cards(self) -> t.Dict:
        raise NotImplementedError('get_cards not implemented')

    @abstractmethod
    def add_card(self, card_number: str) -> t.Dict:
        raise NotImplementedError('add_card not implemented')

    @abstractmethod
    def delete_card(self, card_number: str) -> t.Dict:
        raise NotImplementedError('delete_card not implemented')

    @abstractmethod
    def get_ibans(self) -> t.Dict:
        raise NotImplementedError('get_ibans not implemented')

    @abstractmethod
    def add_iban(self, iban: str) -> t.Dict:
        raise NotImplementedError('add_iban not implemented')

    @abstractmethod
    def delete_iban(self, iban: str) -> t.Dict:
        raise NotImplementedError('delete_iban not implemented')

    @abstractmethod
    def get_wallets(self, asset: str) -> t.Dict:
        raise NotImplementedError('get_wallets not implemented')

    @abstractmethod
    def get_balances(self, asset: str = None) -> t.Dict:
        raise NotImplementedError('get_balances not implemented')

    @abstractmethod
    def get_available_balance(self, asset: str) -> float:
        raise NotImplementedError('get_available_balance not implemented')

    @abstractmethod
    def get_fees(self, symbol: str = None) -> t.Dict:
        raise NotImplementedError('get_fees not implemented')

    @abstractmethod
    def create_order(
            self, symbol: str, side: str, type: str, quantity: float, price: float = None, client_id: str = None
    ) -> t.Dict:
        raise NotImplementedError('create_order not implemented')

    @abstractmethod
    def order_market(self, symbol: str, side: str, quantity: float, client_id: str = None) -> t.Dict:
        raise NotImplementedError('order_market not implemented')

    @abstractmethod
    def order_limit(self, symbol: str, side: str, quantity: float, price: float, client_id: str = None) -> t.Dict:
        raise NotImplementedError('order_limit not implemented')

    @abstractmethod
    def order_market_buy(self, symbol: str, quantity: float, client_id: str = None) -> t.Dict:
        raise NotImplementedError('order_market_buy not implemented')

    @abstractmethod
    def order_market_sell(self, symbol: str, quantity: float, client_id: str = None) -> t.Dict:
        raise NotImplementedError('order_market_sell not implemented')

    @abstractmethod
    def order_limit_buy(self, symbol: str, quantity: float, price: float, client_id: str = None) -> t.Dict:
        raise NotImplementedError('order_limit_buy not implemented')

    @abstractmethod
    def order_limit_sell(self, symbol: str, quantity: float, price: float, client_id: str = None) -> t.Dict:
        raise NotImplementedError('order_limit_sell not implemented')

    @abstractmethod
    def cancel_order(self, order_id: str) -> t.Dict:
        raise NotImplementedError('cancel_order not implemented')

    @abstractmethod
    def get_open_orders(self, symbol: str = None, side: str = None) -> t.Dict:
        raise NotImplementedError('get_open_orders not implemented')

    @abstractmethod
    def get_user_recent_trades(self, symbol: str = None, side: str = None, active: bool = None) -> t.Dict:
        raise NotImplementedError('get_user_recent_trades not implemented')

    @abstractmethod
    def get_order_status(self, order_id: str) -> t.Dict:
        raise NotImplementedError('get_order_status not implemented')

    @abstractmethod
    def withdraw(
            self, coin: str, network: str, amount: float,
            address: str, client_unique_id: str, memo: t.Optional[str] = None
    ) -> t.Dict:
        raise NotImplementedError('withdraw not implemented')

    @abstractmethod
    def close_connection(self):
        raise NotImplementedError('close_connection not implemented')



import typing as t
import json

from requests import Response
from aiohttp import ClientResponse


class APIException(Exception):
    def __init__(self, response: t.Union[Response, ClientResponse], status_code: int, text: str):
        self.code = 0
        try:
            json_res = json.loads(text)
        except ValueError:
            self.message = 'Invalid JSON error message from Wallex: {}'.format(response.text)
        else:
            self.code = json_res['code']
            self.message = json_res['message']
            self.result = json_res['result']

        self.status_code = status_code
        self.response = response
        self.request = getattr(response, 'request', None)

    def __str__(self):  # pragma: no cover
        return 'APIError(code=%s): %s | %s' % (self.code, self.message, self.result)


class RequestException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return 'RequestException: %s' % self.message


class OrderException(Exception):
    def __init__(self, field: str, message: str):
        self.field = field
        self.message = message

    def __str__(self):
        return 'OrderException(code=%s): %s' % (self.field, self.message)



from inspect import currentframe as cf
import typing as t

import requests


from .exceptions import *
from ._package_data import __version__, __author__


class Wallex:
    """
    Wallex API wrapper.
    """

    MARKET_ORDER = 'market'
    LIMIT_ORDER = 'limit'
    STOP_LIMIT_ORDER = 'stop_limit'
    STOP_MARKET_ORDER = 'stop_market'

    BUY_ORDER = 'buy'
    SELL_ORDER = 'sell'

    def __init__(self, token: t.Optional[str] = None, requests_params: t.Optional[t.Dict] = None):
        self.__base_url = "https://api.wallex.ir/v1/"
        self.__requests_params = requests_params
        self.__token = token

        self.__headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
        }

        if self.__token:
            self.__headers['x-api-key'] = self.__token

        if self.__requests_params:
            if 'headers' in self.__requests_params.keys():
                self.__headers.update(self.__requests_params.get('headers'))

        return

    @staticmethod
    def __get_func_args(all_locals: t.Dict, remove_keys: t.Optional[t.List[str]] = None) -> t.Dict:
        """
        Get function arguments.

        :param all_locals: All locals
        :type all_locals: dict

        :return: Function arguments
        :rtype: dict
        """

        if remove_keys is None:
            remove_keys = ['self']

        if 'self' not in remove_keys:
            remove_keys.append('self')

        for i in remove_keys:
            try:
                all_locals.pop(i)
            except KeyError:
                pass

        return all_locals

    def __make_headers(self, func_name: str, auth: t.Optional[bool] = True) -> t.Dict:
        """
        Make headers.

        :param func_name: Function name
        :type func_name: str

        :param auth: Authorization (optional)
        :type auth: bool

        :return: Headers
        :rtype: dict
        """

        f_args = self.__get_func_args(locals())
        headers = self.__headers.copy()

        if auth is True:
            if self.__token is None:
                raise TokenException(
                    func_name,
                    'access_token is None',
                    f_args=f_args
                )
            headers.update({
                'Authorization': f'Bearer {self.__token}'
            })

        return headers

    @staticmethod
    def __validate_response(f_name: str, response: requests.Response) -> bool:
        r_json = response.json()
        _ = r_json.get('success') is True or str(r_json.get('success')).lower() == 'true'

        if _ is False:
            raise InvalidResponse(
                f_name,
                'invalid response',
                response=response
            )

        return _

    def _process_response(
            self,
            func_name: str,
            response: requests.Response,
            additional: t.Optional[t.Dict] = None,
            load_result: bool = True
    ) -> t.Dict:
        """
        Check response for exceptions.

        :param response: Response
        :type response: requests.Response

        :param func_name: Function name
        :type func_name: str

        :raises: WallexExceptions

        :return: None
        :rtype: None
        """

        if additional is not None:
            additional.update(self.__get_func_args(locals()))

        if 200 <= response.status_code < 300:
            try:
                r_json: t.Dict = response.json()
            except Exception as e:
                raise JSONDecodingError(
                    func_name=func_name,
                    message='Could not decode JSON',
                    response=response,
                )

        else:
            raise StatusCodeError(
                func_name,
                f'invalid status code',
                response=response,
                additional=additional
            )

        self.__validate_response(func_name, response)

        if load_result is False:
            return r_json

        return r_json.get('result')

    def _request(
            self, func_name: str, method: str, url: str, auth: t.Optional[bool] = False,
            params: t.Optional[t.Dict] = None, data: t.Optional[t.Dict] = None, json_data: t.Optional[t.Dict] = None,
    ) -> requests.Response:
        """
        Make a request to the Bitpin API.

        :param method: HTTP method
        :type method: str

        :param url: URL
        :type url: str

        :param auth: Whether to use authentication (optional)
        :type auth: bool

        :param params: Query parameters (optional)
        :type params: dict

        :param data: Request body (optional)
        :type data: dict

        :param json_data: Request body (optional)
        :type json_data: dict

        :param func_name: Function name
        :type func_name: str

        :return: Response
        :rtype: requests.Response
        """

        f_args = self.__get_func_args(locals())
        headers = self.__make_headers(func_name, auth)

        try:
            response = requests.request(
                method,
                self.__base_url + url,
                headers=headers,
                params=params,
                json=json_data,
                data=data
            )

            return response
        except Exception as e:
            raise RequestsExceptions(
                func_name,
                e,
                f_args=f_args
            )

    def set_token(self, token: str) -> str:
        """
        Set token

        :param token: Token
        :type token: str

        :return: Token
        :rtype: str
        """

        self.__token = token
        return self.__token

    def markets_stats(self, symbol: t.Optional[str] = None) -> t.Dict:
        """
        Get markets stat.

        :return: Markets stat
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        try:
            response = self._request(
                f_name,
                'GET',
                'markets'
            )
        except Exception as e:
            raise RequestsExceptions(f_name, e)

        resp = self._process_response(f_name, response)

        if symbol is not None:
            _ = resp.get('symbols').get(symbol.upper())
            if _ is None:
                raise InvalidInputs(
                    f_name,
                    f'Invalid symbol, Symbol Not Found',
                    symbol=symbol,
                )
            return _

        return resp

    def orderbook(self, symbol: str) -> t.Dict[str, t.List[t.Dict[str, t.Any]]]:
        """
        Get orderbook.

        :param symbol: Symbol
        :type symbol: str

        :return: Orderbook
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        try:
            response = self._request(
                f_name,
                'GET',
                f'depth?symbol={symbol.upper()}'
            )
        except Exception as e:
            raise RequestsExceptions(f_name, e)

        return self._process_response(f_name, response)

    def balances(self, asset: t.Optional[str] = None) -> t.Dict:
        """
        Get balances.

        :param asset: Asset (optional)
        :type asset: str

        :return: Balances
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        try:
            response = self._request(
                f_name,
                'GET',
                'account/balances',
                auth=True
            )
        except Exception as e:
            raise RequestsExceptions(f_name, e)

        resp = self._process_response(f_name, response)

        if asset is not None:
            return resp.get('balances').get(asset.upper())

        return resp

    def available_balance(self, asset: str) -> float:
        """
        Get available balances.

        :return: Available balances
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        try:
            balances = self.balances(asset)
            return float(balances.get('value')) - float(balances.get('locked'))

        except Exception as e:
            raise WallexExceptions(f_name, e)

    def create_order(
            self,
            symbol: str,
            side: str,
            quantity: t.Union[float, int],
            price: t.Union[float, int],
            order_type: str,
            stop_price: t.Optional[t.Union[float, int]] = None,
            client_order_id: t.Optional[str] = None
    ) -> t.Dict:
        """
        Create order.

        :param symbol: Symbol
        :type symbol: str

        :param side: Side
        :type side: str

        :param quantity: Amount
        :type quantity: str

        :param price: Price
        :type price: str

        :param order_type: Order type
        :type order_type: str

        :param stop_price: Stop price (optional)
        :type stop_price: str

        :param client_order_id: Client order ID (optional)
        :type client_order_id: str

        :return: Order
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        data = {
            "price": str(price),
            "quantity": str(quantity),
            "side": side.lower(),
            "symbol": symbol.upper(),
            "type": order_type.lower(),
        }

        if client_order_id is not None:
            data.update({'client_id': client_order_id})

        if order_type.lower() == self.STOP_LIMIT_ORDER or order_type.lower() == self.STOP_MARKET_ORDER:
            if stop_price is None:
                raise InvalidInputs(
                    f_name,
                    f'Stop price is required for {order_type} order',
                    order_type=order_type,
                    stop_price=stop_price
                )
            data.update({'stop_price': str(stop_price)})

        try:
            response = self._request(
                f_name,
                'POST',
                'account/orders',
                auth=True,
                json_data=data
            )
        except Exception as e:
            raise RequestsExceptions(f_name, e)

        return self._process_response(f_name, response)

    def order_market(self, symbol: str, side: str, quantity: float, client_order_id: t.Optional[str] = None) -> t.Dict:
        """
        Create market order.

        :param symbol: Symbol
        :type symbol: str

        :param side: Side
        :type side: str

        :param quantity: Amount
        :type quantity: str

        :param client_order_id: Client order ID (optional)
        :type client_order_id: str

        :return: Order
        :rtype: dict
        """

        return self.create_order(
            symbol=symbol,
            side=side,
            quantity=quantity,
            price=0,
            order_type=self.MARKET_ORDER,
            client_order_id=client_order_id
        )

    def order_market_buy(self, symbol: str, quantity: float, client_order_id: t.Optional[str] = None) -> t.Dict:
        """
        Create market buy order.

        :param symbol: Symbol
        :type symbol: str

        :param quantity: Amount
        :type quantity: str

        :param client_order_id: Client order ID (optional)
        :type client_order_id: str

        :return: Order
        :rtype: dict
        """

        return self.order_market(
            symbol=symbol,
            side=self.BUY_ORDER,
            quantity=quantity,
            client_order_id=client_order_id
        )

    def order_market_sell(self, symbol: str, quantity: float, client_order_id: t.Optional[str] = None) -> t.Dict:
        """
        Create market sell order.

        :param symbol: Symbol
        :type symbol: str

        :param quantity: Amount
        :type quantity: str

        :param client_order_id: Client order ID (optional)
        :type client_order_id: str

        :return: Order
        :rtype: dict
        """

        return self.order_market(
            symbol=symbol,
            side=self.SELL_ORDER,
            quantity=quantity,
            client_order_id=client_order_id
        )

    def order_limit(
            self, symbol: str, side: str, quantity: float, price: float,
            stop_price: t.Optional[t.Union[float, int]] = None, client_order_id: t.Optional[str] = None
    ) -> t.Dict:
        """
        Create limit order.

        :param symbol: Symbol
        :type symbol: str

        :param side: Side
        :type side: str

        :param quantity: Amount
        :type quantity: str

        :param price: Price
        :type price: str

        :param stop_price: Stop price (optional)
        :type stop_price: str

        :param client_order_id: Client order ID (optional)
        :type client_order_id: str

        :return: Order
        :rtype: dict
        """

        return self.create_order(
            symbol=symbol,
            side=side,
            quantity=quantity,
            price=price,
            order_type=self.LIMIT_ORDER,
            stop_price=stop_price,
            client_order_id=client_order_id
        )

    def order_limit_buy(
            self, symbol: str, quantity: float, price: float,
            stop_price: t.Optional[t.Union[float, int]] = None, client_order_id: t.Optional[str] = None
    ) -> t.Dict:
        """
        Create limit buy order.

        :param symbol: Symbol
        :type symbol: str

        :param quantity: Amount
        :type quantity: str

        :param price: Price
        :type price: str

        :param stop_price: Stop price (optional)
        :type stop_price: str

        :param client_order_id: Client order ID (optional)
        :type client_order_id: str

        :return: Order
        :rtype: dict
        """

        return self.order_limit(
            symbol=symbol,
            side=self.BUY_ORDER,
            quantity=quantity,
            price=price,
            stop_price=stop_price,
            client_order_id=client_order_id
        )

    def order_limit_sell(
            self, symbol: str, quantity: float, price: float, client_order_id: t.Optional[str] = None
    ) -> t.Dict:
        """
        Create limit sell order.

        :param symbol: Symbol
        :type symbol: str

        :param quantity: Amount
        :type quantity: str

        :param price: Price
        :type price: str

        :param client_order_id: Client order ID (optional)
        :type client_order_id: str

        :return: Order
        :rtype: dict
        """

        return self.order_limit(
            symbol=symbol,
            side=self.SELL_ORDER,
            quantity=quantity,
            price=price,
            client_order_id=client_order_id
        )

    def cancel_order(self, order_id: str) -> t.Dict:
        """
        Cancel order.

        :param order_id: Order ID
        :type order_id: str

        :return: Order
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        data = {
            "clientOrderId": order_id
        }

        try:
            response = self._request(
                f_name,
                'DELETE',
                f'account/orders',
                auth=True,
                json_data=data
            )
        except Exception as e:
            raise RequestsExceptions(f_name, e)

        return self._process_response(f_name, response)

    def open_orders(
            self, symbol: t.Optional[str] = None, side: t.Optional[str] = None, page: int = 1, per_page: int = 200
    ) -> t.Dict:
        """
        Get open orders.

        :param symbol: Symbol (optional)
        :type symbol: str

        :param side: Side (optional)
        :type side: str

        :param page: Page (optional)
        :type page: int

        :param per_page: Per page (optional)
        :type per_page: int

        :return: Open orders
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        url = f'account/openOrders'

        params = {
            'page': page,
            'per_page': per_page
        }

        if symbol is not None:
            params.update({'symbol': symbol.upper()})

        try:
            response = self._request(
                f_name,
                'GET',
                url,
                auth=True,
                params=params
            )
        except Exception as e:
            raise RequestsExceptions(f_name, e)

        resp = self._process_response(f_name, response, load_result=False)

        if side is not None:
            all_orders = resp.get('result').get('orders')
            for order in all_orders:
                if order.get('side').lower() != side.lower():
                    all_orders.remove(order)

        return resp

    def user_recent_trades(
            self, symbol: t.Optional[str] = None, side: t.Optional[str] = None, active: t.Optional[bool] = None,
            page: int = 1, per_page: int = 200
    ) -> t.Dict:
        """
        Get user recent trades.

        :param symbol: Symbol (optional)
        :type symbol: str

        :param side: Side (optional)
        :type side: str

        :param active: Active (optional)
        :type active: bool

        :param page: Page (optional)
        :type page: int

        :param per_page: Per page (optional)
        :type per_page: int

        :return: User recent trades
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        params = {
            'page': page,
            'per_page': per_page
        }

        if symbol:
            params.update({'symbol': symbol.upper()})
        if side:
            params.update({'side': side.lower()})
        if active:
            params.update({'active': active})

        try:
            response = self._request(
                f_name,
                'GET',
                'account/trades',
                auth=True,
                params=params
            )
        except Exception as e:
            raise RequestsExceptions(f_name, e)

        return self._process_response(f_name, response, load_result=False)

    def order_status(self, order_id: str) -> t.Dict:
        """
        Get order status.

        :param order_id: Order ID
        :type order_id: str

        :return: Order status
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        try:
            response = self._request(
                f_name,
                'GET',
                f'account/orders/{order_id}',
                auth=True,
            )
        except Exception as e:
            raise RequestsExceptions(f_name, e)

        return self._process_response(f_name, response)

    def withdraw(self, coin: str, network: str, amount: t.Union[int, float],
                 address: str, client_unique_id: str, memo: t.Optional[str] = None):
        """
        Withdraw.

        :param coin: Coin
        :type coin: str

        :param client_unique_id: Client unique ID
        :type client_unique_id: str

        :param network: Network
        :type network: str

        :param amount: Amount
        :type amount: str

        :param address: Address
        :type address: str

        :param memo: Memo (optional)
        :type memo: str

        :return: Withdraw
        :rtype: dict
        """

        f_name = cf().f_code.co_name

        data = {
            "coin": coin.upper(),
            "client_unique_id": str(client_unique_id),
            "network": network,
            "value": str(amount),
            "wallet_address": address,
        }
        if memo:
            data.update({'memo': memo})

        try:
            response = self._request(
                f_name,
                'POST',
                'account/crypto-withdrawal',
                auth=True,
                json_data=data
            )
        except Exception as e:
            raise RequestsExceptions(f_name, e)

        return self._process_response(f_name, response)

    def __str__(self):
        return f'Wallex REST API Client, By {__author__}, Version {__version__}'

    def __repr__(self):
        return f'Wallex REST API Client, By {__author__}, Version {__version__}'



import typing as t
import requests
import aiohttp
import asyncio

from ..enums import Resolution
from .base import BaseClient
from ..exceptions import RequestException, APIException


__all__ = [
    'Client',
    'AsyncClient'
]


class Client(BaseClient):
    def __init__(
            self, api_key: t.Optional[str] = None, requests_params: t.Optional[t.Dict[str, t.Any]] = None,
    ):

        super().__init__(api_key, requests_params)

    def _init_session(self) -> requests.Session:

        headers = self._get_headers()

        session = requests.session()
        session.headers.update(headers)
        return session

    def _request(self, method, uri: str, signed: bool, **kwargs):

        kwargs = self._get_request_kwargs(method, signed, **kwargs)

        self.response = getattr(self.session, method)(uri, **kwargs)
        return self._handle_response(self.response)

    @staticmethod
    def _handle_response(response: requests.Response):
        if not (200 <= response.status_code < 300):
            raise APIException(response, response.status_code, response.text)
        try:
            return response.json()
        except ValueError:
            raise RequestException('Invalid Response: %s' % response.text)

    def _request_api(
            self, method, path: str, signed: bool = False, version=BaseClient.PUBLIC_API_VERSION, **kwargs
    ):
        uri = self._create_api_uri(path, version)
        return self._request(method, uri, signed, **kwargs)

    def _get(self, path, signed=False, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        return self._request_api('get', path, signed, version, **kwargs)

    def _post(self, path, signed=False, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        return self._request_api('post', path, signed, version, **kwargs)

    def _put(self, path, signed=False, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        return self._request_api('put', path, signed, version, **kwargs)

    def _delete(self, path, signed=False, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        return self._request_api('delete', path, signed, version, **kwargs)

    def get_market_stats(self, symbol: str = None) -> t.Dict:
        result = self._get('markets')

        if symbol is not None:
            data = self._pick(result.get('result').get('symbols'), symbol)
            result['result']['symbols'] = data

        return result

    def get_currencies(self, currency: str = None) -> t.Dict:
        result = self._get('currencies')

        if currency is not None:
            data = self._pick(result.get('result'), currency)
            result['result'] = data

        return result

    def get_currencies_stats(self, currency=None) -> t.Dict:
        result = self._get('currencies/stats')

        if currency is not None:
            data = self._pick(result.get('result'), 'key', currency)
            result['result'] = data

        return result

    def get_orderbook(self, symbol: str) -> t.Dict:
        return self._get('depth', params={'symbol': symbol})

    def get_recent_trades(self, symbol: str = 'None', **kwargs) -> t.Dict:
        return self._get('trades', params={'symbol': symbol}, **kwargs)

    def get_ohlc_data(
            self, symbol: str = None, resolution: Resolution = None, from_date: int = None, to_date: int = None
    ) -> t.Dict:
        return self._get('udf/history', params={
            'symbol': symbol,
            'resolution': resolution.value if resolution is not None else None,
            'from': from_date,
            'to': to_date
        })

    def get_profile(self) -> t.Dict:
        return self._get('account/profile', signed=True)

    def get_cards(self) -> t.Dict:
        return self._get('account/card-numbers', signed=True)

    def add_card(self, card_number: str) -> t.Dict:
        return self._post('account/card-numbers', signed=True, json={'card_number': card_number})

    def delete_card(self, card_number: str) -> t.Dict:
        return self._delete(f'account/card-numbers/{card_number}', signed=True)

    def get_ibans(self) -> t.Dict:
        return self._get('account/ibans', signed=True)

    def add_iban(self, iban: str) -> t.Dict:
        return self._post('account/ibans', signed=True, json={'iban': iban})

    def delete_iban(self, iban: str) -> t.Dict:
        return self._delete(f'account/ibans/{iban}', signed=True)

    def get_wallets(self, asset: str) -> t.Dict:
        return self._get(f'account/wallets/{asset}', signed=True)

    def get_balances(self, asset: str = None) -> t.Dict:
        result = self._get(f'account/balances', signed=True)

        if asset is not None:
            data = self._pick(result.get('result').get('balances'), asset)
            result['result']['balances'] = data

        return result

    def get_available_balance(self, asset: str) -> float:
        result = self.get_balances(asset)
        result = result.get('result').get('balances').get(asset.upper())
        return float(result.get('value')) - float(result.get('locked'))

    def get_fees(self, symbol: str = None) -> t.Dict:
        result = self._get('account/fee', signed=True)

        if symbol is not None:
            data = self._pick(result.get('result'), symbol)
            result['result'] = data

        return result

    def create_order(
            self, symbol: str, side: str, type: str, quantity: float, price: float = None, client_id: str = None
    ) -> t.Dict:
        return self._post('account/orders', signed=True, json=self._get_kwargs(locals(), del_nones=True))

    def order_market(self, symbol: str, side: str, quantity: float, client_id: str = None) -> t.Dict:
        return self.create_order(
            symbol=symbol, side=side, type=self.ORDER_TYPE_MARKET, quantity=quantity, client_id=client_id
        )

    def order_limit(self, symbol: str, side: str, quantity: float, price: float, client_id: str = None) -> t.Dict:
        return self.create_order(
            symbol=symbol, side=side, type=self.ORDER_TYPE_LIMIT, quantity=quantity, client_id=client_id, price=price
        )

    def order_market_buy(self, symbol: str, quantity: float, client_id: str = None) -> t.Dict:
        return self.order_market(
            symbol=symbol, side=self.SIDE_BUY, quantity=quantity, client_id=client_id
        )

    def order_market_sell(self, symbol: str, quantity: float, client_id: str = None) -> t.Dict:
        return self.order_market(
            symbol=symbol, side=self.SIDE_SELL, quantity=quantity, client_id=client_id
        )

    def order_limit_buy(self, symbol: str, quantity: float, price: float, client_id: str = None) -> t.Dict:
        return self.order_limit(
            symbol=symbol, side=self.SIDE_BUY, quantity=quantity, price=price, client_id=client_id
        )

    def order_limit_sell(self, symbol: str, quantity: float, price: float, client_id: str = None) -> t.Dict:
        return self.order_limit(
            symbol=symbol, side=self.SIDE_SELL, quantity=quantity, price=price, client_id=client_id
        )

    def cancel_order(self, order_id: str) -> t.Dict:
        return self._delete(f'account/orders', signed=True, json={'clientOrderId': order_id})

    def get_open_orders(self, symbol: str = None, side: str = None) -> t.Dict:
        result = self._get('account/openOrders', signed=True)

        if symbol is not None:
            data = self._pick(result.get('result').get('orders'), "symbol", symbol)
            result['result']['orders'] = data

        if side is not None:
            data = self._pick(result.get('result').get('orders'), "side", side)
            result['result']['orders'] = data

        return result

    def get_user_recent_trades(self, symbol: str = None, side: str = None, active: bool = None) -> t.Dict:
        return self._get('account/trades', signed=True, params={
            'symbol': symbol,
            'side': side,
            'active': active
        })

    def get_order_status(self, order_id: str) -> t.Dict:
        return self._get(f'account/orders/{order_id}', signed=True)

    def withdraw(
            self, coin: str, network: str, amount: float,
            address: str, client_unique_id: str, memo: t.Optional[str] = None
    ) -> t.Dict:
        return self._post('account/crypto-withdrawal', signed=True, json=self._get_kwargs(locals(), del_nones=True))

    def close_connection(self):
        if self.session:
            self.session.close()

    def __del__(self):
        self.close_connection()


class AsyncClient(BaseClient):
    def __init__(
            self,
            api_key: t.Optional[str] = None,
            requests_params: t.Optional[t.Dict[str, t.Any]] = None,
            loop: t.Optional[asyncio.AbstractEventLoop] = None
    ):

        self.loop = loop or asyncio.get_event_loop()
        super().__init__(api_key, requests_params)

    @classmethod
    async def create(
            cls,
            api_key: t.Optional[str] = None,
            requests_params: t.Optional[t.Dict[str, t.Any]] = None,
            loop: t.Optional[asyncio.AbstractEventLoop] = None
    ) -> 'AsyncClient':

        return cls(api_key, requests_params, loop)

    def __aenter__(self):
        return self

    def __aexit__(self, exc_type, exc_val, exc_tb):
        self.close_connection()
        return False

    def _init_session(self) -> aiohttp.ClientSession:
        session = aiohttp.ClientSession(
            loop=self.loop,
            headers=self._get_headers()
        )
        return session

    async def _request(self, method, uri: str, signed: bool, **kwargs):

        kwargs = self._get_request_kwargs(method, signed, **kwargs)

        async with getattr(self.session, method)(uri, **kwargs) as response:
            self.response = response
            return await self._handle_response(response)

    @staticmethod
    async def _handle_response(response: aiohttp.ClientResponse):
        if not str(response.status).startswith('2'):
            raise APIException(response, response.status, await response.text())
        try:
            return await response.json()
        except ValueError:
            txt = await response.text()
            raise RequestException(f'Invalid Response: {txt}')

    async def _request_api(self, method, path, signed=False, version=BaseClient.PUBLIC_API_VERSION, **kwargs):
        uri = self._create_api_uri(path, version)
        return await self._request(method, uri, signed, **kwargs)

    async def _get(self, path, signed=False, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        return await self._request_api('get', path, signed, version, **kwargs)

    async def _post(self, path, signed=False, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        return await self._request_api('post', path, signed, version, **kwargs)

    async def _put(self, path, signed=False, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        return await self._request_api('put', path, signed, version, **kwargs)

    async def _delete(self, path, signed=False, version=BaseClient.PUBLIC_API_VERSION, **kwargs) -> t.Dict:
        return await self._request_api('delete', path, signed, version, **kwargs)

    async def get_market_stats(self, symbol: str = None) -> t.Dict:
        result = await self._get('markets')

        if symbol is not None:
            data = self._pick(result.get('result').get('symbols'), symbol)
            result['result']['symbols'] = data

        return result

    async def get_currencies(self, currency: str = None) -> t.Dict:
        result = await self._get('currencies')

        if currency is not None:
            data = self._pick(result.get('result'), currency)
            result['result'] = data

        return result

    async def get_currencies_stats(self, currency=None) -> t.Dict:
        result = await self._get('currencies/stats')

        if currency is not None:
            data = self._pick(result.get('result'), 'key', currency)
            result['result'] = data

        return result

    async def get_orderbook(self, symbol: str) -> t.Dict:
        return await self._get('depth', params={'symbol': symbol})

    async def get_recent_trades(self, symbol: str = 'None', **kwargs) -> t.Dict:
        return await self._get('trades', params={'symbol': symbol}, **kwargs)

    async def get_ohlc_data(
            self, symbol: str = None, resolution: Resolution = None, from_date: int = None, to_date: int = None
    ) -> t.Dict:
        return await self._get('udf/history', params={
            'symbol': symbol,
            'resolution': resolution.value if resolution is not None else None,
            'from': from_date,
            'to': to_date
        })

    async def get_profile(self) -> t.Dict:
        return await self._get('account/profile', signed=True)

    async def get_cards(self) -> t.Dict:
        return await self._get('account/card-numbers', signed=True)

    async def add_card(self, card_number: str) -> t.Dict:
        return await self._post('account/card-numbers', signed=True, json={'card_number': card_number})

    async def delete_card(self, card_number: str) -> t.Dict:
        return await self._delete(f'account/card-numbers/{card_number}', signed=True)

    async def get_ibans(self) -> t.Dict:
        return await self._get('account/ibans', signed=True)

    async def add_iban(self, iban: str) -> t.Dict:
        return await self._post('account/ibans', signed=True, json={'iban': iban})

    async def delete_iban(self, iban: str) -> t.Dict:
        return await self._delete(f'account/ibans/{iban}', signed=True)

    async def get_wallets(self, asset: str) -> t.Dict:
        return await self._get(f'account/wallets/{asset}', signed=True)

    async def get_balances(self, asset: str = None) -> t.Dict:
        result = await self._get(f'account/balances', signed=True)

        if asset is not None:
            data = self._pick(result.get('result').get('balances'), asset)
            result['result']['balances'] = data

        return result

    async def get_available_balance(self, asset: str) -> float:
        result = await self.get_balances(asset)

        result = result.get('result').get('balances').get(asset.upper())

        return float(result.get('value')) - float(result.get('locked'))

    async def get_fees(self, symbol: str = None) -> t.Dict:
        result = await self._get('account/fee', signed=True)

        if symbol is not None:
            data = self._pick(result.get('result'), symbol)
            result['result'] = data

        return result

    async def create_order(
            self, symbol: str, side: str, type: str, quantity: float, price: float = None, client_id: str = None
    ) -> t.Dict:
        return await self._post('account/orders', signed=True, json=self._get_kwargs(locals(), del_nones=True))

    async def order_market(self, symbol: str, side: str, quantity: float, client_id: str = None) -> t.Dict:
        return await self.create_order(
            symbol=symbol, side=side, type=self.ORDER_TYPE_MARKET, quantity=quantity, client_id=client_id
        )

    async def order_limit(self, symbol: str, side: str, quantity: float, price: float, client_id: str = None) -> t.Dict:
        return await self.create_order(
            symbol=symbol, side=side, type=self.ORDER_TYPE_LIMIT, quantity=quantity, client_id=client_id, price=price
        )

    async def order_market_buy(self, symbol: str, quantity: float, client_id: str = None) -> t.Dict:
        return await self.order_market(
            symbol=symbol, side=self.SIDE_BUY, quantity=quantity, client_id=client_id
        )

    async def order_market_sell(self, symbol: str, quantity: float, client_id: str = None) -> t.Dict:
        return await self.order_market(
            symbol=symbol, side=self.SIDE_SELL, quantity=quantity, client_id=client_id
        )

    async def order_limit_buy(self, symbol: str, quantity: float, price: float, client_id: str = None) -> t.Dict:
        return await self.order_limit(
            symbol=symbol, side=self.SIDE_BUY, quantity=quantity, price=price, client_id=client_id
        )

    async def order_limit_sell(self, symbol: str, quantity: float, price: float, client_id: str = None) -> t.Dict:
        return await self.order_limit(
            symbol=symbol, side=self.SIDE_SELL, quantity=quantity, price=price, client_id=client_id
        )

    async def cancel_order(self, order_id: str) -> t.Dict:
        return await self._delete(f'account/orders', signed=True, json={'clientOrderId': order_id})

    async def get_open_orders(self, symbol: str = None, side: str = None) -> t.Dict:
        result = await self._get('account/openOrders', signed=True)

        if symbol is not None:
            data = self._pick(result.get('result').get('orders'), "symbol", symbol)
            result['result']['orders'] = data

        if side is not None:
            data = self._pick(result.get('result').get('orders'), "side", side)
            result['result']['orders'] = data

        return result

    async def get_user_recent_trades(self, symbol: str = None, side: str = None, active: bool = None) -> t.Dict:
        return await self._get('account/trades', signed=True, params={
            'symbol': symbol,
            'side': side,
            'active': active
        })

    async def get_order_status(self, order_id: str) -> t.Dict:
        return await self._get(f'account/orders/{order_id}', signed=True)

    async def withdraw(
            self, coin: str, network: str, amount: float,
            address: str, client_unique_id: str, memo: t.Optional[str] = None
    ) -> t.Dict:
        return await self._post(
            'account/crypto-withdrawal', signed=True, json=self._get_kwargs(locals(), del_nones=True)
        )

    async def close_connection(self):
        if self.session:
            assert self.session
            await self.session.close()

    def __del__(self):
        asyncio.get_event_loop().create_task(self.close_connection())



__name__ = "wallex"
__version__ = "0.4.5"
__author__ = "amiwrpremium"
__reason__ = 'OK'
__email__ = "amiwrpremium@gmail.com"
__license__ = "MIT"
__url__ = "https://github.com/amiwrpremium/wallex"
__description__ = "Wallex cryptocurrency exchange python SDK"



import typing as t
from pydantic import BaseModel, Field, root_validator
from datetime import datetime


class ResultInfo(BaseModel):
    page: int
    per_page: int
    count: int
    total_count: int


class _Response(BaseModel):
    result: t.Dict[str, BaseModel]
    message: str
    success: bool


class MarketStats(_Response):
    class Result(BaseModel):
        class Symbols(BaseModel):
            class Stats(BaseModel):
                class Direction(BaseModel):
                    SELL: int
                    BUY: int

                bidPrice: float
                askPrice: float
                twentyFourHourChannel: t.Union[str, float] = Field(None, alias='24h_ch')
                sevenDayChannel: t.Union[str, float] = Field(None, alias='7d_ch')
                twentyFourHourVolume: t.Union[str, float] = Field(None, alias='24h_volume')
                sevenDayVolume: t.Union[str, float] = Field(None, alias='7d_volume')
                twentyFourHourQuoteVolume: t.Union[str, float] = Field(None, alias='24h_quoteVolume')
                twentyFourHourHighPrice: t.Union[str, float] = Field(None, alias='24h_highPrice')
                twentyFourHourLowPrice: t.Union[str, float] = Field(None, alias='24h_lowPrice')
                lastPrice: t.Union[float, str]
                lastQty: t.Union[float, str]
                lastTradeSide: str
                bidVolume: float
                askVolume: float
                bidCount: int
                askCount: int
                direction: Direction

            symbol: str
            baseAsset: str
            baseAssetPrecision: int
            quoteAsset: str
            quotePrecision: int
            faName: str
            faBaseAsset: str
            faQuoteAsset: str
            stepSize: int
            tickSize: int
            minQty: float
            minNotional: float
            stats: Stats
            createdAt: datetime

        symbols: t.Dict[str, Symbols]

    result: Result


class Currencies(_Response):
    class Result(BaseModel):
        class TransactionFee(BaseModel):
            percent: float
            max_fee: float

        class Network(BaseModel):
            id: int
            name: str
            message: t.Optional[str]
            type: str
            deposit_availability: str
            withdrawal_availability: str
            deposit_unavailability_reason: t.Optional[str]
            withdrawal_unavailability_reason: t.Optional[str]
            min_confirmation: int
            transaction_fee: float
            min_withdrawal_value: float
            min_deposit_value: float

        key: str
        name: str
        name_en: str
        type: str
        base: bool
        precision: int
        message: t.Optional[str]
        deposit_availability: str
        withdrawal_availability: str
        updated_at: t.Optional[datetime]
        transaction_fee: t.Optional[TransactionFee]
        min_withdrawal_value: t.Optional[float]
        network: t.Optional[t.List[Network]]

    result: t.Dict[str, Result]
    provider: str
    result_info: ResultInfo


class CurrenciesStats(_Response):
    class Result(BaseModel):
        key: str
        name: str
        name_en: str
        rank: t.Optional[int]
        dominance: t.Optional[float]
        volume_24h: t.Optional[float]
        market_cap: t.Optional[float]
        ath: t.Optional[float]
        ath_change_percentage: t.Optional[float]
        ath_date: t.Optional[datetime]
        price: t.Optional[float]
        daily_high_price: t.Optional[float]
        daily_low_price: t.Optional[float]
        weekly_high_price: t.Optional[float]
        weekly_low_price: t.Optional[float]
        percent_change_1h: t.Optional[float]
        percent_change_24h: t.Optional[float]
        percent_change_7d: t.Optional[float]
        percent_change_14d: t.Optional[float]
        percent_change_30d: t.Optional[float]
        percent_change_60d: t.Optional[float]
        percent_change_200d: t.Optional[float]
        percent_change_1y: t.Optional[float]
        price_change_24h: t.Optional[float]
        price_change_7d: t.Optional[float]
        price_change_14d: t.Optional[float]
        price_change_30d: t.Optional[float]
        price_change_60d: t.Optional[float]
        price_change_200d: t.Optional[float]
        price_change_1y: t.Optional[float]
        max_supply: t.Optional[float]
        total_supply: t.Optional[float]
        circulating_supply: t.Optional[float]
        created_at: t.Optional[datetime]
        updated_at: t.Optional[datetime]

    result: t.List[Result]
    provider: str
    result_info: ResultInfo


class Orderbook(_Response):
    class Result(BaseModel):
        class Order(BaseModel):
            price: float
            quantity: float
            sum: float

        ask: t.List[Order]
        bid: t.List[Order]

    result: Result


class RecentTrades(_Response):
    class Result(BaseModel):
        class Trade(BaseModel):
            symbol: str
            quantity: float
            price: float
            sum: float
            isBuyOrder: bool
            timestamp: datetime

        latestTrades: t.List[Trade]

    result: Result

#
# class OHLC(_Response):
#     raise NotImplementedError

    # class Result(BaseModel):
    #     class OHLC(BaseModel):
    #         open: t.List[float] = Field(None, alias='o')
    #         high: t.List[float] = Field(None, alias='h')
    #         low: t.List[float] = Field(None, alias='l')
    #         close: t.List[float] = Field(None, alias='c')
    #         volume: t.List[float] = Field(None, alias='v')
    #         timestamp: t.List[datetime] = Field(None, alias='t')
    #
    #     ohlc: t.List[OHLC]
    #
    # result: Result


setting_sample = {
    'result': {
        'tracking_id': 244283,
        'first_name': 'امیرمهدی',
        'last_name': 'عرفانی زاده',
        'national_code': '0024****103',
        'face_image': None,
        'birthday': '2001-07-31T00:00:00Z',
        'address': {
            'city': '',
            'country': '',
            'location': 'شهرآرا، کوچه 38، پلاک 21, واحد 3',
            'province': '',
            'postal_code': '',
            'house_number': ''
        },
        'phone_number': {
            'area_code': '021',
            'main_number': '88265241'
        },
        'mobile_number': '0939****814',
        'verification': 'VERIFIED',
        'email': 'am**********@g****.com',
        'invite_code': 'BnAvR',
        'avatar': None,
        'commission': 25,
        'settings': {
            'theme': 'light',
            'mode': 'pro',
            'order_submit_confirm': False,
            'order_delete_confirm': True,
            'default_mode': True,
            'favorite_markets': [],
            'choose_trading_type': True,
            'coin_deposit': True,
            'coin_withdraw': True,
            'money_deposit': True,
            'money_withdraw': True,
            'logins': True,
            'trade': True,
            'api_key_expiration': True,
            '0': 'manual_deposit',
            'notification': {
                'email': {
                    'is_enable': True,
                    'actions': {
                        'coin_deposit': {'is_enable': True, 'label': 'واریز کوین'},
                        'coin_withdraw': {'is_enable': True, 'label': 'برداشت کوین'},
                        'money_deposit': {'is_enable': True, 'label': 'واریز تومان'},
                        'money_withdraw': {'is_enable': True, 'label': 'برداشت تومان'},
                        'logins': {'is_enable': True, 'label': 'فعالیتها'},
                        'api_key_expiration': {'is_enable': True, 'label': 'labels.profile_settings.api_key_expiration'},
                        'manual_deposit': {'is_enable': True, 'label': 'labels.profile_settings.manual_deposit'}
                    },
                    'label': 'ایمیل'
                },
                'announcement': {
                    'is_enable': True,
                    'actions': {
                        'coin_deposit': {'is_enable': True, 'label': 'واریز کوین'},
                        'coin_withdraw': {'is_enable': True, 'label': 'برداشت کوین'},
                        'money_deposit': {'is_enable': True, 'label': 'واریز تومان'},
                        'money_withdraw': {'is_enable': True, 'label': 'برداشت تومان'},
                        'logins': {'is_enable': True, 'label': 'فعالیتها'},
                        'trade': {'is_enable': True, 'label': 'معامله'},
                        'api_key_expiration': {'is_enable': True, 'label': 'labels.profile_settings.api_key_expiration'},
                        'manual_deposit': {'is_enable': True, 'label': 'labels.profile_settings.manual_deposit'},
                        'price_alert': {'is_enable': True, 'label': 'labels.profile_settings.price_alert'}},
                    'label': 'اطلاعیه داخل سایت(اپلیکیشن)'
                },
                'push': {
                    'is_enable': True,
                    'actions': {
                        'coin_deposit': {'is_enable': True, 'label': 'واریز کوین'},
                        'coin_withdraw': {'is_enable': True, 'label': 'برداشت کوین'},
                        'money_deposit': {'is_enable': True, 'label': 'واریز تومان'},
                        'money_withdraw': {'is_enable': True, 'label': 'برداشت تومان'},
                        'logins': {'is_enable': True, 'label': 'فعالیتها'},
                        'trade': {'is_enable': True, 'label': 'معامله'},
                        'api_key_expiration': {'is_enable': True, 'label': 'labels.profile_settings.api_key_expiration'},
                        'manual_deposit': {'is_enable': True,'label': 'labels.profile_settings.manual_deposit'},
                        'price_alert': {'is_enable': True,'label': 'labels.profile_settings.price_alert'}
                    },
                    'label': 'Push Notification'
                }
            }
        },
        'status': {
            'first_name': 'ACCEPTED',
            'last_name': 'ACCEPTED',
            'national_code': 'ACCEPTED',
            'national_card_image': 'ACCEPTED',
            'face_image': 'UNFILLED',
            'face_video': 'UNFILLED',
            'birthday': 'ACCEPTED',
            'address': 'ACCEPTED',
            'phone_number': 'ACCEPTED',
            'mobile_number': 'ACCEPTED',
            'email': 'ACCEPTED'
        },
        'kyc_info': {
            'details': {
                'mobile_activation': True,
                'personal_info': True,
                'financial_info': True,
                'phone_number': True,
                'national_card': True,
                'face_recognition': True,
                'admin_approval': True
            },
            'level': 2
        },
        'meta': {
            'disabled_features': [
                'BOT_USER_CRYPTO_WITHDRAWAL_EXTRA_VERIFICATION_STEP',
                'quick trade',
                'QUICK_TRADE',
                'REWARDS_HUB'
            ]
        }
    },
    'message': 'The operation was successful',
    'success': True
}


class Profile(_Response):
    class Result(BaseModel):
        class Address(BaseModel):
            city: str
            country: str
            location: str
            province: str
            postal_code: str
            house_number: str

        class PhoneNumber(BaseModel):
            area_code: str
            main_number: str

        class Settings(BaseModel):
            class Notification(BaseModel):
                class Email(BaseModel):
                    class Actions(BaseModel):
                        class Action(BaseModel):
                            is_enable: bool
                            label: str

                        coin_deposit: Action
                        coin_withdraw: Action
                        money_deposit: Action
                        money_withdraw: Action
                        logins: Action
                        api_key_expiration: Action
                        manual_deposit: Action

                    is_enable: bool
                    actions: Actions
                    label: str

                class AnnouncementAndPush(BaseModel):
                    class Action(BaseModel):
                        is_enable: bool
                        label: str

                    coin_deposit: Action
                    coin_withdraw: Action
                    money_deposit: Action
                    money_withdraw: Action
                    logins: Action
                    trade: Action
                    api_key_expiration: Action
                    manual_deposit: Action
                    price_alert: Action

                    is_enable: bool
                    actions: Action
                    label: str

                email: Email
                announcement: AnnouncementAndPush
                push: AnnouncementAndPush

            theme: str
            mode: str
            order_submit_confirm: bool
            order_delete_confirm: bool
            default_mode: bool
            favorite_markets: t.List[str]
            choose_trading_type: bool
            coin_deposit: bool
            coin_withdraw: bool
            money_deposit: bool
            money_withdraw: bool
            logins: bool
            trade: bool
            api_key_expiration: bool
            zero: str = Field(None, alias='0')
            notification: Notification

        tracking_id: int
        first_name: str
        last_name: str
        national_code: str
        face_image: t.Optional[str]
        birthday: datetime
        address: Address
        phone_number: PhoneNumber
        mobile_number: str
        verification: str
        email: str
        invite_code: t.Optional[str]
        avatar: t.Optional[str]
        commission: int
        settings: Settings

    result: Result


class Wallets(_Response):
    class Result(BaseModel):
        class CoinType(BaseModel):
            key: str
            name: str
            name_en: str
            type: str
            deposit_availability: str = Field(None, enum=['ENABLE', 'DISABLE'])
            withdrawal_availability: str = Field(None, enum=['ENABLE', 'DISABLE'])
            deposit_unavailability_reason: t.Optional[str]
            withdrawal_unavailability_reason: t.Optional[str]

        class Wallet(BaseModel):
            class Network(BaseModel):
                id: int
                name: str
                message: t.Optional[str]
                type: str
                deposit_availability: str = Field(None, enum=['ENABLE', 'DISABLE'])
                withdrawal_availability: str = Field(None, enum=['ENABLE', 'DISABLE'])
                deposit_unavailability_reason: t.Optional[str]
                withdrawal_unavailability_reason: t.Optional[str]

            address: str
            memo: t.Optional[str]
            min_confirmation: int
            transaction_fee: float
            min_withdrawal_value: float
            min_deposit_value: float

        coin_type: CoinType
        wallets: t.Dict[str, Wallet]

    result: Result
    provider: str


class Balances(_Response):
    class Result(BaseModel):
        class Balance(BaseModel):
            asset: str
            faName: str
            fiat: bool
            value: float
            locked: float
            free: float = None

            @root_validator(pre=True)
            def _calculate_free_balance(cls, values):
                total = float(values.get('value'))
                locked = float(values.get('locked'))

                free = values.get('free')

                if free is None:
                    free = total - locked
                values['free'] = float(free)
                return values

        balances: t.Dict[str, Balance]

    result: Result


class Fees(_Response):
    class Result(BaseModel):
        makerFeeRate: t.Optional[float]
        takerFeeRate: t.Optional[float]
        recent_days_sum: t.Optional[float]

    result: t.Dict[str, Result]


class Order(BaseModel):
    class Result(BaseModel):
        class Fills(BaseModel):
            price: float
            quantity: float
            fee: float
            feeCoefficient: float
            feeAsset: str
            timestamp: datetime
            symbol: str
            sum: float
            makerFeeCoefficient: float
            takerFeeCoefficient: float
            isBuyer: bool

            side: str = None

            @root_validator(pre=True)
            def _find_side(cls, values):
                if values.get('isBuyer'):
                    values['side'] = 'buy'
                else:
                    values['side'] = 'sell'
                return values

        symbol: str = Field(None, alias='symbol')
        type: str = Field(None, alias='type')
        side: str = Field(None, alias='side')
        client_order_id: str = Field(None, alias='clientOrderId')
        transact_time: datetime = Field(None, alias='transactTime')
        price: float = Field(None, alias='price')
        orig_qty: float = Field(None, alias='origQty')
        executed_sum: float = Field(None, alias='executedSum')
        executed_qty: float = Field(None, alias='executedQty')
        executed_price: float = Field(None, alias='executedPrice')
        sum: float = Field(None, alias='sum')
        executed_percent: float = Field(None, alias='executedPercent')
        status: str = Field(None, alias='status')
        active: bool = Field(None, alias='active')
        fills: t.List[Fills] = Field(None, alias='fills')
        created_at: t.Optional[datetime] = Field(None, alias='created_at')

    result: Result
    # TODO add enum for status


class OpenOrders(_Response):
    class Result(BaseModel):
        orders: t.List[Order.Result]

    result: Result



from distutils.core import setup
from setuptools import find_packages

setup(
    name='wallex',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    version='0.4.5',
    license='MIT',
    description='Wallex cryptocurrency exchange python sdk',
    author='amiwrpremium',
    author_email='amiwrpremium@gmail.com',
    url='https://github.com/amiwrpremium/wallex',
    keywords=['wallex', 'crypto', 'exchange', 'API', "SDK"],
    install_requires=[
        'requests',
        'simplejson',
        'deprecation',
        'python-socketio==4.6.1',
        'python-engineio==3.14.2',
        'websocket-client==1.3.2',
        'aiohttp[speedups]',
        'pydantic'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



import typing as t


import socketio
import engineio


from .exceptions import WebsocketNotConnected
from ._package_data import __version__, __author__


if socketio.__version__ < '4.0.0' or socketio.__version__ >= '5.0.0':
    raise RuntimeError('socketio version should be > 4.0.0 and <= 5.0.0')
if engineio.__version__ < '3.0.0' or engineio.__version__ >= '4.0.0':
    raise RuntimeError('engineio version should be > 2.0.0 and <= 3.0.0')


class WallexWebsocket:
    """
    Wallex Websocket class.
    """

    def __init__(self, url: str = 'https://api.wallex.ir/socket.io',
                 socketio_params: t.Optional[t.Dict[str, t.Any]] = None,
                 debug: t.Optional[bool] = False):
        """
        Wallex Websocket class constructor.

        :param url: Websocket URL
        :type url: str

        :param socketio_params: SocketIO parameters (optional)
        :type socketio_params: t.Dict[str, t.Any]

        :param debug: Enable debug mode (optional)
        :type debug: t.Optional[bool]

        :return: Wallex Websocket class instance
        :rtype: WallexWebsocket
        """
        self._sio: socketio.Client = socketio.Client(**socketio_params or {})
        self._url: str = url
        self._debug: bool = debug

        self._is_connected: bool = False

    def __exc_if_not_connected(self, f_name: str):
        """
        Raise exception if not connected.

        :param f_name: Function name
        :type f_name: str

        :return: None
        """

        if not self._is_connected:
            raise WebsocketNotConnected(f_name, 'Websocket is not connected')

    def __debug_print(self, msg: str, *args, **kwargs):
        """
        Print debug message if debug mode is enabled.

        :param msg: Message to print
        :type msg: str

        :param args: Arguments to pass to print
        :type args: t.Any

        :param kwargs: Keyword arguments to pass to print
        :type kwargs: t.Any

        :return: None
        """

        message = str(msg)

        if args and len(args) > 0:
            message += f" | {args}"

        if kwargs and len(kwargs) > 0:
            message += f" | {kwargs}"

        if self._debug:
            print(message)

    def connect(self, url: t.Optional[str] = None, params: t.Optional[t.Dict[str, t.Any]] = None):
        """
        Connect to a websocket.

        :param url: Websocket URL (optional)
        :type url: t.Optional[str]

        :param params: SocketIO parameters (optional)
        :type params: t.Optional[t.Dict[str, t.Any]]

        :return: None
        """

        _url = url or self._url
        _params = params or {}
        if 'transports' not in _params.keys():
            _params['transports'] = ['websocket']

        self._sio.connect(_url, **_params)
        self._is_connected = True
        self.__debug_print('Connection established')

    def disconnect(self):
        """
        Disconnect from a websocket.

        :return: None
        """

        self.__exc_if_not_connected('disconnect')
        self._sio.disconnect()
        self._is_connected = False
        self.__debug_print('Connection closed')

    def subscribe(self, channel: str):
        """
        Subscribe to a channel

        :param channel: Channel name
        :type channel: str

        :return: None
        """

        self.__exc_if_not_connected('subscribe')
        self.emit('subscribe', data={"channel": channel})
        self.__debug_print(f'Subscribed to channel: {channel}')

    def unsubscribe(self, channel: str):
        """
        Unsubscribe from a channel

        :param channel: Channel name
        :type channel: str

        :return: None
        """

        self.__exc_if_not_connected('unsubscribe')
        self.emit('unsubscribe', data={"channel": channel})
        self.__debug_print(f'Unsubscribed from channel: {channel}')

    def wait(self):
        """
        Wait for a message.

        :return: None
        """

        self.__exc_if_not_connected('wait')
        self._sio.wait()

    def emit(self, event: str, data: t.Dict[str, t.Any]):
        """
        Emit an event.

        :param event: Event name
        :type event: str

        :param data: Event data
        :type data: t.Dict[str, t.Any]

        :return: None
        """

        self.__exc_if_not_connected('emit')
        self._sio.emit(event, data)
        self.__debug_print(f'Emitted event', event=event, data=data)

    def on(self, event: str, callback: t.Callable[[str, t.Dict[str, t.Any]], None]):
        """
        Register a callback for an event.

        :param event: Event name
        :type event: str

        :param callback: Callback function
        :type callback: t.Callable[[str, t.Dict[str, t.Any]], None]

        :return: None
        """

        self.__exc_if_not_connected('on')
        self._sio.on(event, callback)
        self.__debug_print(f'Registered callback for event', event=event)

    def on_connect(self, callback: t.Callable[[], None]):
        """
        Register a callback for connection.

        :param callback: Callback function
        :type callback: t.Callable[[], None]

        :return: None
        """

        self.__exc_if_not_connected('on_connect')
        self.on('connect', callback)

    def on_disconnect(self, callback: t.Callable[[], None]):
        """
        Register a callback for disconnection.

        :param callback: Callback function
        :type callback: t.Callable[[], None]

        :return: None
        """

        self.__exc_if_not_connected('on_disconnect')
        self.on('disconnect', callback)

    def on_error(self, callback: t.Callable[[str, t.Dict[str, t.Any]], None]):
        """
        Register a callback for error.

        :param callback: Callback function
        :type callback: t.Callable[[str, t.Dict[str, t.Any]], None]

        :return: None
        """

        self.__exc_if_not_connected('on_error')
        self.on('error', callback)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def __str__(self):
        return f'Wallex Websocket Client, By {__author__}, Version {__version__}'

    def __repr__(self):
        return f'Wallex Websocket API Client, By {__author__}, Version {__version__}'
